// Simulated storage device backed by a file

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <unistd.h>
#include <inttypes.h>
#include <fcntl.h>
#include "flash_api.h"
#include "bench.h"
#include <sys/mman.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <sys/stat.h>

struct backend * backend_open(const char *filename, bool truncate) {
    struct backend *ba = calloc(sizeof(struct backend), 1);
    if (!ba) {
        printf("[backend] alloc err\n");
        return NULL;
    }

    ba->path = filename;

    errno = 0;
    ba->fd = open(filename, O_RDWR | O_CREAT | (truncate ? O_TRUNC : 0));
    if (ba->fd < 0) {
        perror("Open file");
        free(ba);
        return false;
    }

    fchmod(ba->fd, S_IRUSR | S_IWUSR);

    errno = 0;

    // ensure the file has the right size
    ftruncate(ba->fd, CHIP_SIZE);

    ba->mapped_size = CHIP_SIZE;

    // this can be used to rule out bugs due to OS caching
    ba->mapped = mmap((caddr_t)0, ba->mapped_size, PROT_READ | PROT_WRITE, MAP_SHARED, ba->fd, 0);
    //ba->mapped = calloc(ba->mapped_size, 1);

    if (!ba->mapped) {
        perror("Map memory");
        goto fail;
    }

    ba->bench = bench_init(BLOCK_COUNT);
    if (!ba->bench) {
        printf("Err alloc bench!\n");
        goto fail;
    }

    if (truncate) {
        backend_wipe(ba);
    }

    return ba;

fail:
    //if (ba->mapped) munmap(ba->mapped, ba->mapped_size);
    if (ba->bench) bench_free(ba->bench);
    if (ba->fd) close(ba->fd);
    free(ba);

    return NULL;
}


void backend_wipe(struct backend *b) {
    assert(b);
    assert(b->mapped);

    memset((void *)b->mapped, 0xFF, b->mapped_size);
}

void backend_flush(struct backend *b) {
    assert(b);
    assert(b->mapped);
    fsync(b->fd);
}

void backend_close(struct backend *b) {
    if (b->mapped) {
        munmap((void *)b->mapped, b->mapped_size);
        b->mapped = NULL;
    }

    if (b->bench) {
        bench_free(b->bench);
        b->bench = NULL;
    }

    if (b->fd) {
        close(b->fd);
        b->fd = 0;
    }

    free(b);
}

bool backend_read(struct backend *b, uint32_t addr, uint8_t *data, uint16_t len) {
    assert(b);
    assert(b->mapped);

    if (addr + len <= b->mapped_size) {
        memcpy(data, (void *)(b->mapped + addr), len);
        return true;
    } else {
        printf("[backend] Read exceeds chip bounds! 0x%"PRIx32" + %d", addr, len);
        return false;
    }
}

bool backend_write(struct backend *b, uint32_t addr, uint8_t *data, uint16_t len) {
    assert(b);
    assert(b->mapped);

    if (addr + len <= b->mapped_size) {
        volatile uint8_t *wp = b->mapped + addr;
        for (; len > 0; len--, wp++, data++) {
            *wp = *wp & *data; // NOR effect - 1 can become 0, but 0 can never go back to 1 unless the block is erased
        }
        return true;
    } else {
        printf("[backend] Write exceeds chip bounds! 0x%"PRIx32" + %d", addr, len);
        return false;
    }
}

bool backend_erase(struct backend *b, uint32_t addr, uint32_t len) {
    assert(b);
    assert(b->mapped);

    if (addr + len <= b->mapped_size) {
        memset(b->mapped + addr, 0xFF, len);
        return true;
    } else {
        printf("[backend] Erase exceeds chip bounds! 0x%"PRIx32" + %d", addr, len);
        return false;
    }
}
