// Simulated storage device backed by a file

#ifndef FSIM_BACKEND_H
#define FSIM_BACKEND_H

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

/**
 * This is the backend struct, exposed as public for easy access
 * to the memory mapped area and bench.
 */
struct backend {
    const char *path; //!< Backing file path

    int fd; //!< open filedes
    volatile uint8_t *mapped; //!< memory area where the file is mapped (in its entirety)
    size_t mapped_size; //!< size of the memmapped area

    struct bench *bench; //!< Active bench associated with this backend
};

/**
 * Open the backend
 *
 * @param filename - file to open
 * @param truncate - truncate the file (start with em "empty" disk - all 0xFF)
 * @return pointer to the allocated backend struct
 */
struct backend* __attribute__((warn_unused_result)) backend_open(const char *filename, bool truncate);

/**
 * Safely de-allocate the backend struct.
 * If 'bench' is still set, it is freed using `bench_free()`.
 * The struct is freed and the caller must not use it after calling this method.
 *
 * @param b - pointer to the backend struct
 */
void backend_close(struct backend *b);

/**
 * Attempt to flush the backend file to disk.
 *
 * @param b
 */
void backend_flush(struct backend *b);

/**
 * Read from the backend file at a given offset
 *
 * @param b - pointer to the backend struct
 * @param addr - address to read at
 * @param data - data buffer
 * @param len - read length (buffer capacity)
 * @return success
 */
bool __attribute__((warn_unused_result)) backend_read(struct backend *b, uint32_t addr, uint8_t *data, uint16_t len);

/**
 * Write to the backend file at a given offset.
 * NOR effects are applied - i.e. the write is AND-ed with the original content
 *
 * @param b - pointer to the backend struct
 * @param addr - address to write at
 * @param data - data buffer
 * @param len - data length
 * @return success
 */
bool __attribute__((warn_unused_result)) backend_write(struct backend *b, uint32_t addr, uint8_t *data, uint16_t len);

/**
 * Set a file area to 0xFF (erased)
 *
 * @param b - pointer to the backend struct
 * @param addr - address to erase at (this MUST be at a block boundary)
 * @param len - erased bytes count (this MUST be a whole block)
 * @return success
 */
bool __attribute__((warn_unused_result)) backend_erase(struct backend *b, uint32_t addr, uint32_t len);
void backend_wipe(struct backend *b);

#endif /* FSIM_BACKEND_H */
