#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <inttypes.h>
#include <assert.h>

#include "bench.h"
#include "utils.h"

struct block_info {
    uint8_t erased;
    uint8_t written;
};

struct bench {
    uint64_t io_time_ns; // nanoseconds spent doing IO ops (simulated)
    uint32_t total_erase_num; // total number of block erases
    uint32_t block_count; // block count
    struct block_info *blocks; // erase counters
};


#define MAX_BLOCK_ERASE_NUM 255
struct bench * bench_init(uint32_t block_count) {
    struct bench *bench = calloc(sizeof(struct bench), 1);
    if (!bench) {
        printf("[bench] alloc err\n");
        return NULL;
    }

    //printf("[bench] init\n");
    bzero(bench, sizeof(struct bench));
    bench->blocks = calloc(block_count, sizeof(struct block_info));
    bench->block_count = block_count;
    return bench;
}

void bench_free(struct bench * bench) {
    free(bench);
}

void bench_clear(struct bench * bench) {
    assert(bench->blocks);
    //printf("[bench] clear counters\n");

    bench->io_time_ns = 0;
    bench->total_erase_num = 0;
    bzero(bench->blocks, bench->block_count * sizeof(struct block_info));
}

uint64_t bench_get_io_time(struct bench * bench) {
    return bench->io_time_ns;
}

void bench_add_io_time(struct bench * bench, uint64_t ns) {
    bench->io_time_ns += ns;
}

void bench_block_erased(struct bench * bench, uint32_t num) {
    assert(bench->blocks);
    printf("[bench] block %d erased\n", num);

    bench->total_erase_num++;
    if (num < bench->block_count) {
        if (bench->blocks[num].erased < MAX_BLOCK_ERASE_NUM) {
            bench->blocks[num].erased++;
        }
    } else {
        printf("[bench] ERR: Erase block out of range: %"PRIu32"\n", num);
    }
}

void bench_block_written(struct bench * bench, uint32_t num) {
    assert(bench->blocks);
    //printf("[bench] block %d written\n", num);

    if (num < bench->block_count) {
        if (bench->blocks[num].written < MAX_BLOCK_ERASE_NUM) {
            bench->blocks[num].written++;
        }
    } else {
        printf("[bench] ERR: Write block out of range: %"PRIu32"\n", num);
    }
}

void bench_show_stats(struct bench * bench, bool colors) {
    printf("\n[bench] --- stats (%"PRIu32" blocks) ---\n", bench->block_count);
    printf("[bench] Total blocks erased: %"PRIu32"\n", bench->total_erase_num);

    print_elapsed("[bench] Time spent in IO", bench->io_time_ns);

    printf("[bench] Block erase counts:\n");

    for (int i = 0; i < bench->block_count / 32; i++) {
        printf("[bench] %3d: ", i*32);
        for (int j = 0; j < 32 && ((i*32+j)<bench->block_count); j++) {
            int count = bench->blocks[i*32 + j].erased;

            if (colors) {
                const char *color =
                    count == 0 ? "90" // gray
                    : count < 5 ? "32" // green
                    : count < 10 ? "33" // yellow
                    : count < 15 ? "31" // red
                    : "41;37"; // red bg

                const char *extra = "";
                if (bench->blocks[i*32 + j].written > 0) {
                    extra = "44;";
                    if (count==0) color = "37";
                }

                printf("\x1b[%s%sm%3d\x1b[m", extra, color, count);
            } else {
                printf("%3d", count);
            }
        }
        printf("\n");
    }
    printf("\n");
}
