#include "fs_spiffs.h"
#include "flash_api.h"
#include "bench.h"

#include <spiffs.h>
#include <inttypes.h>
#include <assert.h>

static spiffs g_spiffs_fs_;
spiffs *g_spiffs_fs = &g_spiffs_fs_;
struct backend *g_spiffs_backend;

static uint8_t s_spiffs_work_buf[2 * PAGE_SIZE];
#define S_SPIFFS_FD_SPACE_SIZE 128
static uint8_t s_spiffs_fd_space[S_SPIFFS_FD_SPACE_SIZE];
#define S_SPIFFS_CACHE_SIZE (32*PAGE_SIZE)
static uint8_t s_spiffs_cache[S_SPIFFS_CACHE_SIZE];

static s32_t spi_read_spiffs(u32_t addr, u32_t size, u8_t *dst);
static s32_t spi_write_spiffs(u32_t addr, u32_t size, u8_t *src);
static s32_t spi_erase_spiffs(u32_t addr, u32_t size);
static void spiffs_check_callback_impl(spiffs_check_type type, spiffs_check_report report, u32_t arg1, u32_t arg2);

const static spiffs_config cfg = {
    // physical read function
    .hal_read_f = spi_read_spiffs,
    // physical write function
    .hal_write_f = spi_write_spiffs,
    // physical erase function
    .hal_erase_f = spi_erase_spiffs,

    // physical size of the spi flash
    .phys_size = CHIP_SIZE,
    // physical offset in spi flash used for spiffs,
    // must be on block boundary
    .phys_addr = 0,
    // physical size when erasing a block
    .phys_erase_block = BLOCK_SIZE,

    // logical size of a block, must be on physical
    // block size boundary and must never be less than
    // a physical block
    .log_block_size = BLOCK_SIZE,
    // logical size of a page, must be at least
    // log_block_size / 8
    .log_page_size = PAGE_SIZE,
};

bool fs_init_spiffs(const char *file, bool wipe) {
    if (NULL == (g_spiffs_backend = backend_open(file, wipe))) {
        printf("backend open failed.\n");
        return false;
    }

    return true;
}

uint64_t fs_iotime_spiffs(void) {
    return bench_get_io_time(g_spiffs_backend->bench);
}

bool fs_mount_spiffs(void) {
    // must not be already mounted to mount
    assert(!SPIFFS_mounted(pSPIFFS));

    /**
     * Initializes the file system dynamic parameters and mounts the filesystem.
     * If SPIFFS_USE_MAGIC is enabled the mounting may fail with SPIFFS_ERR_NOT_A_FS
     * if the flash does not contain a recognizable file system.
     * In this case, SPIFFS_format must be called prior to remounting.
     * @param fs            the file system struct
     * @param config        the physical and logical configuration of the file system
     * @param work          a memory work buffer comprising 2*config->log_page_size
     *                      bytes used throughout all file system operations
     * @param fd_space      memory for file descriptors
     * @param fd_space_size memory size of file descriptors
     * @param cache         memory for cache, may be null
     * @param cache_size    memory size of cache
     * @param check_cb_f    callback function for reporting during consistency checks
     */
    int32_t rv = SPIFFS_mount(pSPIFFS,
                              (spiffs_config*) &cfg,
                              s_spiffs_work_buf,
                              s_spiffs_fd_space, S_SPIFFS_FD_SPACE_SIZE,
                              s_spiffs_cache, S_SPIFFS_CACHE_SIZE,
                              spiffs_check_callback_impl
    );

    if (rv != SPIFFS_OK) {
        printf("[int] Error mounting SPIFFS: %"PRIi32"\n", rv);
        return false;
    }

    return true;
}

void fs_umount_spiffs(void) {
    SPIFFS_unmount(pSPIFFS);
    backend_flush(g_spiffs_backend);
}

/**
 * SPIFFS read function
 *
 * @param addr
 * @param size
 * @param dst
 * @return spiffs error code, 0 = OK
 */
static s32_t spi_read_spiffs(u32_t addr, u32_t size, u8_t *dst) {
    if (!flash_read_data(g_spiffs_backend, addr, dst, size)) {
        SPIFFS_DBG("[integ] ERR: Read at addr %"PRIu32", size %"PRIu32" failed.", addr, size);
        return SPIFFS_ERR_INTERNAL;
    }
    return SPIFFS_OK;
}

/**
 * SPIFFS write function
 *
 * @param addr
 * @param size
 * @param src
 * @return spiffs error code, 0 = OK
 */
static s32_t spi_write_spiffs(u32_t addr, u32_t size, u8_t *src) {
    if (!flash_write_data(g_spiffs_backend, addr, src, size)) {
        SPIFFS_DBG("[integ] ERR: Write to addr %"PRIu32", size %"PRIu32" failed.", addr, size);
        return SPIFFS_ERR_INTERNAL;
    }
    return SPIFFS_OK;
}

/**
 * SPIFFS block erase function
 *
 * @param addr
 * @param size
 * @return spiffs error code, 0 = OK
 */
static s32_t spi_erase_spiffs(u32_t addr, u32_t size) {
    if (size != BLOCK_SIZE) {
        SPIFFS_DBG("[integ] ERR: Attempt to erase non-block size %"PRIu32, size);
        return SPIFFS_ERR_ERASE_FAIL;
    }

    if (addr % BLOCK_SIZE != 0) {
        SPIFFS_DBG("[integ] ERR: Attempted unaligned block erase: start at %"PRIx32, addr);
        return SPIFFS_ERR_ERASE_FAIL;
    }

    if (!flash_erase_block(g_spiffs_backend, addr)) {
        SPIFFS_DBG("[integ] ERR: Erase at addr %"PRIu32", size %"PRIu32" failed.", addr, size);
        return SPIFFS_ERR_INTERNAL;
    }
    return SPIFFS_OK;
}

static void spiffs_check_callback_impl(spiffs_check_type type, spiffs_check_report report, u32_t arg1, u32_t arg2) {
    static const char* type_s[] = {
        [SPIFFS_CHECK_LOOKUP] = "LOOKUP",
        [SPIFFS_CHECK_INDEX] = "INDEX",
        [SPIFFS_CHECK_PAGE] = "PAGE",
    };
    static const char* report_s[] = {
        [SPIFFS_CHECK_PROGRESS] = "PROGRESS",
        [SPIFFS_CHECK_ERROR] = "ERROR",
        [SPIFFS_CHECK_FIX_INDEX] = "FIX_INDEX",
        [SPIFFS_CHECK_FIX_LOOKUP] = "FIX_LOOKUP",
        [SPIFFS_CHECK_DELETE_ORPHANED_INDEX] = "DELETE_ORPHANED_INDEX",
        [SPIFFS_CHECK_DELETE_PAGE] = "DELETE_PAGE",
        [SPIFFS_CHECK_DELETE_BAD_FILE] = "DELETE_BAD_FILE"
    };
    SPIFFS_DBG("[integ] %s: %s: %"PRIu32", %"PRIu32, type_s[type], report_s[report], arg1, arg2);
}
