#include <stdint.h>

#include <spiffs.h>
#include <spiffs_config.h>
#include <assert.h>
#include <inttypes.h>

#include "fs_spiffs.h"
#include "backend.h"
#include "bench.h"
#include "minmax.h"
#include "utils.h"

int test_spiffs()
{
    int32_t rv;

#define NEGCHECK(op) do { \
    rv = (op); \
    if(rv<0) { \
        printf("SPIFFS r/w failed at %d - code %d\n", __LINE__, SPIFFS_errno(pSPIFFS)); \
        return 1; \
    } \
} while(0)

#define RVCHECK(op) do { \
    rv = (op); \
    if(rv!=0) { \
        printf("SPIFFS err %"PRIi32" at %d\n", rv, __LINE__); \
        return 1; \
    } \
} while(0)

    if (!fs_init_spiffs("spiffs.img", true)) {
        printf("Err init spiffs\n");
        return 1;
    }

    // special dance needed by SPIFFS to format a disk
    fs_mount_spiffs();
    fs_umount_spiffs();
    RVCHECK(SPIFFS_format(pSPIFFS));

    bench_clear(g_spiffs_backend->bench);

#define BUF_SIZE 256
    char buf[BUF_SIZE];
    uint64_t t;

    t = fs_iotime_spiffs();
    fs_mount_spiffs();
    print_elapsed("SPIFFS mount empty", fs_iotime_spiffs() - t);

    // create a file
    t = fs_iotime_spiffs();
    int fd = SPIFFS_open(pSPIFFS, "my_file.txt", SPIFFS_O_WRONLY | SPIFFS_O_CREAT, 0);
    const char *s = "This is a text line.";
    NEGCHECK(SPIFFS_write(pSPIFFS, fd, (void*)s, strlen(s) + 1)); // with the zero
    RVCHECK(SPIFFS_close(pSPIFFS, fd));
    print_elapsed("Simple file create & write", fs_iotime_spiffs() - t);

    // read back the text file
    t = fs_iotime_spiffs();
    fd = SPIFFS_open(pSPIFFS, "my_file.txt", SPIFFS_O_RDONLY, 0);
    NEGCHECK(SPIFFS_read(pSPIFFS, fd, buf, 21));
    RVCHECK(SPIFFS_close(pSPIFFS, fd));
    print_elapsed("Simple file open & read", fs_iotime_spiffs() - t);

    // check equality
    assert(0 == strcmp(s, buf));

    // a growing log
    t = fs_iotime_spiffs();
    fd = SPIFFS_open(pSPIFFS, "log_file.txt", SPIFFS_O_WRONLY | SPIFFS_O_CREAT, 0);
    for (int i = 0; i < 100000; i++) {
        snprintf(buf, BUF_SIZE, "This is a log file entry number #%d (0x%x)\n", i, i);
        buf[BUF_SIZE-1] = 0;

        NEGCHECK(SPIFFS_write(pSPIFFS, fd, (void*)buf, strlen(buf)));
    }
    RVCHECK(SPIFFS_close(pSPIFFS, fd));
    print_elapsed("Huge log create & writed", fs_iotime_spiffs() - t);

    t = fs_iotime_spiffs();
    fd = SPIFFS_open(pSPIFFS, "log_file.txt", SPIFFS_O_RDONLY, 0);
    spiffs_stat stat;
    RVCHECK(SPIFFS_fstat(pSPIFFS, fd, &stat));
    uint32_t length = stat.size;

    printf("Log file length is %"PRIu32"\n", length);

    while (length > 0) {
        int read = SPIFFS_read(pSPIFFS, fd, buf, MIN(length, BUF_SIZE));
        NEGCHECK(read);
        //fwrite(buf, read, 1, stdout);
        length -= read;
    }
    RVCHECK(SPIFFS_close(pSPIFFS, fd));
    print_elapsed("Huge log read", fs_iotime_spiffs() - t);

    fs_umount_spiffs();

    t = fs_iotime_spiffs();
    fs_mount_spiffs();
    print_elapsed("SPIFFS mount with content", fs_iotime_spiffs() - t);

    // create a file
    t = fs_iotime_spiffs();
    fd = SPIFFS_open(pSPIFFS, "other.txt", SPIFFS_O_WRONLY | SPIFFS_O_CREAT, 0);
    s = "here is another very small file";
    NEGCHECK(SPIFFS_write(pSPIFFS, fd, (void*)s, strlen(s) + 1)); // with the zero
    RVCHECK(SPIFFS_close(pSPIFFS, fd));
    print_elapsed("small file 2 create&write", fs_iotime_spiffs() - t);

    t = fs_iotime_spiffs();
    RVCHECK(SPIFFS_remove(pSPIFFS, "my_file.txt"));
    print_elapsed("delete small file 1", fs_iotime_spiffs() - t);

    t = fs_iotime_spiffs();
    RVCHECK(SPIFFS_remove(pSPIFFS, "other.txt"));
    print_elapsed("delete small file 2", fs_iotime_spiffs() - t);

    t = fs_iotime_spiffs();
    fd = SPIFFS_open(pSPIFFS, "other2.txt", SPIFFS_O_WRONLY | SPIFFS_O_CREAT, 0);
    s = "here is another very small file - second one";
    NEGCHECK(SPIFFS_write(pSPIFFS, fd, (void*)s, strlen(s) + 1)); // with the zero
    RVCHECK(SPIFFS_close(pSPIFFS, fd));
    print_elapsed("small file 3 create&write", fs_iotime_spiffs() - t);

    t = fs_iotime_spiffs();
    RVCHECK(SPIFFS_remove(pSPIFFS, "log_file.txt"));
    print_elapsed("delete the large log file", fs_iotime_spiffs() - t);

    t = fs_iotime_spiffs();
    fd = SPIFFS_open(pSPIFFS, "log_file.txt", SPIFFS_O_WRONLY | SPIFFS_O_CREAT, 0);
    for (int i = 0; i < 100000; i++) {
        snprintf(buf, BUF_SIZE, "This is a log file entry number #%d (0x%x)\n", i, i);
        buf[BUF_SIZE-1] = 0;

        NEGCHECK(SPIFFS_write(pSPIFFS, fd, (void*)buf, strlen(buf)));
    }
    RVCHECK(SPIFFS_close(pSPIFFS, fd));
    print_elapsed("Create a new large log file", fs_iotime_spiffs() - t);

    printf("create&drop tiny file\n");

    // try to stress it by many tiny files
    for(int i = 0; i < 1000; i++) {
        fd = SPIFFS_open(pSPIFFS, "tinyfile.txt", SPIFFS_O_WRONLY | SPIFFS_O_CREAT, 0);
        s = "woop dy doop";
        NEGCHECK(SPIFFS_write(pSPIFFS, fd, (void*)s, strlen(s) + 1)); // with the zero
        RVCHECK(SPIFFS_close(pSPIFFS, fd));
        RVCHECK(SPIFFS_remove(pSPIFFS, "tinyfile.txt"));
    }

    // strange errors with this - seems spiffs does not GC properly, or corrupts its internal state


    printf("create&drop big files\n");
    for(int i = 0; i < 100; i++) {
        bench_show_stats(g_spiffs_backend->bench, true);

        fd = SPIFFS_open(pSPIFFS, "bigfile.txt", SPIFFS_O_WRONLY | SPIFFS_O_CREAT, 0);
        for (int i = 0; i < 10000; i++) {
            snprintf(buf, BUF_SIZE, "This is a log file entry number #%d (0x%x)\n", i, i);
            buf[BUF_SIZE-1] = 0;

            NEGCHECK(SPIFFS_write(pSPIFFS, fd, (void*)buf, strlen(buf)));
        }
        RVCHECK(SPIFFS_close(pSPIFFS, fd));

        RVCHECK(SPIFFS_remove(pSPIFFS, "bigfile.txt"));
    }

    fs_umount_spiffs();

    bench_show_stats(g_spiffs_backend->bench, true);

    backend_close(g_spiffs_backend);

    return 0;
}
